using System;
using RepPP.Toolkit;

namespace RepPP.Toolkit.Web {

    //********************************************************************************
    //
	/// <summary>
	/// Base class for Web Services object generators
	/// </summary>
	//
    //********************************************************************************
	public abstract class WebServicesGenBase {
	    private string      m_strFldPropPrefix;
	    private string      m_strGrpPropPrefix;
	    private string      m_strCrdClassPrefix;
        private string      m_strGrpClassPrefix;
	    private bool        m_bAllPropAsString;
	    private bool        m_bAddLineState;
        /// <summary>Prefix for the generated methods and classes </summary>
        public const string WEB_SERV_METHOD_PREFIX          = "WS_";
        /// <summary>Prefix for the generated methods and classes </summary>
        public const string WEB_SERV_CARD_PREFIX            = "WSC_";
        /// <summary>Prefix for the generated methods and classes </summary>
        public const string WEB_SERV_GROUP_PREFIX           = "WSG_";

        //*********************************************************	    
        //
	    /// <summary>
	    /// ctor:           Class constructor
	    /// </summary>
        //
        //*********************************************************
		public WebServicesGenBase() {
		    CardClassPrefix         = WEB_SERV_CARD_PREFIX;
            GroupClassPrefix        = WEB_SERV_GROUP_PREFIX;
		    FieldPropertyPrefix     = "fld";
		    GroupPropertyPrefix     = "grp";
		    AllPropertiesAsString   = true;
		    AddLineState            = true;
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// CardClassPrefix:    Prefix use for the name of a class representing a card
	    /// </summary>
        //
        //*********************************************************
		public virtual string CardClassPrefix {
		    get {
		        return(m_strCrdClassPrefix);
		    }
		    set {
		        m_strCrdClassPrefix = (value == null) ? "" : value;
		    }
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// GroupClassPrefix:    Prefix use for the name of a class representing a group
	    /// </summary>
        //
        //*********************************************************
		public virtual string GroupClassPrefix {
		    get {
		        return(m_strGrpClassPrefix);
		    }
		    set {
		        m_strGrpClassPrefix = (value == null) ? "" : value;
		    }
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// FieldPropertyPrefix:    Prefix to use to generates property attached to Rep++ field
	    /// </summary>
        //
        //*********************************************************
		public virtual string FieldPropertyPrefix {
		    get {
		        return(m_strFldPropPrefix);
		    }
		    set {
		        m_strFldPropPrefix = (value == null) ? "" : value;
		    }
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// GroupPropertyPrefix:    Prefix to use to generates property attached to Rep++ group
	    /// </summary>
        //
        //*********************************************************
		public virtual string GroupPropertyPrefix {
		    get {
		        return(m_strGrpPropPrefix);
		    }
		    set {
		        m_strGrpPropPrefix = (value == null) ? "" : value;
		    }
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// AllPropertiesAsString:   true if all properties must be of type string.
	    /// </summary>
        //
        //*********************************************************
		public virtual bool AllPropertiesAsString {
		    get {
		        return(m_bAllPropAsString);
		    }
		    set {
		        m_bAllPropAsString = value;
		    }
		}

        //*********************************************************	    
        //
	    /// <summary>
	    /// AddLineState:           true to add a property/field for holding the line state
	    /// </summary>
        //
        //*********************************************************
		public virtual bool AddLineState {
		    get {
		        return(m_bAddLineState);
		    }
		    set {
		        m_bAddLineState = value;
		    }
		}

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateFieldPropName:  Generates the name of a property attached to a Rep++ field.
        /// </summary>
        /// <param name="fld">      Rep++ field</param>
        /// <returns>
        /// Property name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateFieldPropName(RepPP.Field fld) {
            string  strFldName;

            strFldName = fld.PropertyName.Replace("$", "");
            return(FieldPropertyPrefix + strFldName);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateGroupName:      Generates the name of a Rep++ group.
        /// </summary>
        /// <param name="grp">      Rep++ group</param>
        /// <returns>
        /// Property name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateGroupName(RepPP.Group grp) {
            return(grp.ClassName.Replace("$", ""));
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateGroupPropName:  Generates the name of a property attached to a Rep++ group.
        /// </summary>
        /// <param name="grp">      Rep++ group</param>
        /// <returns>
        /// Property name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateGroupPropName(RepPP.Group grp) {
            string  strGrpName;

            strGrpName = GenerateGroupName(grp);
            return(GroupPropertyPrefix + strGrpName);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateCardGroupName:      Generates the name of a Rep++ card group.
        /// </summary>
        /// <param name="cgrp">         Rep++ card group</param>
        /// <returns>
        /// Name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateCardGroupName(RepPP.CardGroup cgrp) {
            return(cgrp.ClassName.Replace("$", ""));
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateCardGroupPropName:  Generates the name of a property attached to a Rep++ card group.
        /// </summary>
        /// <param name="cgrp">         Rep++ card group</param>
        /// <returns>
        /// Property name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateCardGroupPropName(RepPP.CardGroup cgrp) {
            string  strGrpName;

            strGrpName = GenerateCardGroupName(cgrp);
            return(GroupPropertyPrefix + strGrpName);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateCardClassName:  Generates the name of a class attached to a Rep++ card.
        /// </summary>
        /// <param name="crd">      Rep++ Card</param>
        /// <returns>
        /// Class name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateCardClassName(RepPP.Card crd) {
            string  strCrdName;

            strCrdName = crd.ClassName.Replace("$", "");
            return(CardClassPrefix + strCrdName);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateGroupClassName: Generates the name of a class attached to a Rep++ group.
        /// </summary>
        /// <param name="cgrp">     Card group if any</param>
        /// <param name="grp">      Rep++ group</param>
        /// <returns>
        /// Class name
        /// </returns>
        //
        //*********************************************************
        protected virtual string GenerateGroupClassName(RepPP.CardGroup cgrp, RepPP.Group grp) {
            string          strRetVal;
            //RepPP.CardGroup cgrpParent;
            
            if (cgrp != null) {
                strRetVal  = GenerateCardGroupName(cgrp);
            //    cgrpParent = cgrp.Parent as RepPP.CardGroup;
            //    while (cgrpParent != null) {
            //        strRetVal  = GenerateCardGroupName(cgrpParent) + "_" + strRetVal;
            //        cgrpParent = cgrpParent.Parent as RepPP.CardGroup;
            //    }
                strRetVal = GenerateCardClassName(cgrp.Card) + "_" + strRetVal;
            } else {
                strRetVal = GroupClassPrefix + GenerateGroupName(grp);
            }
            return(strRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GenerateGroupBaseClassName:                 Generates the base class name of a class attached 
        ///                                             to a Rep++ card group.
        /// </summary>
        /// <param name="cgrp">                         Card group </param>
        /// <param name="strTypedInstGrpClassPrefix">   Prefix for Typed instance group class</param>
        /// <returns>
        /// Class name
        /// </returns>
        //
        //*********************************************************
        protected string GenerateGroupBaseClassName(RepPP.CardGroup cgrp, string strTypedInstGrpClassPrefix) {
            string  strRetVal;
            string  strGrpName;

            strGrpName = GenerateGroupName(cgrp.Group);
            strRetVal  = strTypedInstGrpClassPrefix + strGrpName + "." + GroupClassPrefix + strGrpName;
            return(strRetVal);
        }

        //*********************************************************	    
        //
	    /// <summary>
	    /// PurifyRemark:               Be sure a remark doesn't contain illegal character
	    /// </summary>
	    /// <param name="strRem">       Remark</param>
	    /// <returns>
	    /// Purified remark
	    /// </returns>
        //
        //*********************************************************
        protected string PurifyRemark(string strRem) {
            string  strRetVal;
            
            strRetVal = strRem.Replace('<', ' ').Replace('>', ' ').Replace("///", "");
            return(strRetVal);
        }

        //*********************************************************	    
        //
	    /// <summary>
	    /// GenerateGroupClass:                       Generates a class to use in Web Service to represent a group 
	    /// </summary>
	    /// <param name="grp">                        Rep++ group</param>
	    /// <param name="iIndent">                    Indentation to use</param>
	    /// <param name="strBaseClass">               Base class</param>
        /// <param name="strTypedInstGrpClassPrefix"> Prefix for Typed instance group class</param>
        //
        //*********************************************************
        public abstract string GenerateGroupClass(RepPP.Group grp, int iIndent, string strBaseClass, string strTypedInstGrpClassPrefix);

        //*********************************************************	    
        //
	    /// <summary>
	    /// GenerateCardClass:                        Generates a class to use in Web Service to represent a card
	    /// </summary>
	    /// <param name="crd">                        Rep++ card</param>
	    /// <param name="iIndent">                    Indentation to use</param>
	    /// <param name="strCardBaseClass">           Card Base class</param>
	    /// <param name="strGroupBaseClass">          Group Base class</param>
        /// <param name="strTypedInstGrpClassPrefix"> Prefix for Typed instance group class</param>
        //
        //*********************************************************
        public abstract string GenerateCardClass(RepPP.Card crd, int iIndent, string strCardBaseClass, string strGroupBaseClass, string strTypedInstGrpClassPrefix);

        //*********************************************************	    
        //
	    /// <summary>
	    /// GenerateGroupSerializationFunctions:Generates serialization function for the group
	    /// </summary>
	    /// <param name="strContainerClassName">Container class name</param>
	    /// <param name="cgrp">                 Rep++ card group</param>
	    /// <param name="grp">                  Rep++ group</param>
	    /// <param name="iIndent">              Indentation to use</param>
	    /// <returns>
	    /// Generated functions
	    /// </returns>
	    //
        //*********************************************************
        public abstract string GenerateGroupSerializationFunctions(string strContainerClassName, RepPP.CardGroup cgrp, RepPP.Group grp, int iIndent);

        //*********************************************************	    
        //
	    /// <summary>
	    /// GenerateCardSerializationFunctions: Generates serialization function for the card
	    /// </summary>
	    /// <param name="strContainerClassName">Container class name</param>
	    /// <param name="crd">                  Rep++ card</param>
	    /// <param name="iIndent">              Indentation to use</param>
	    /// <returns>
	    /// Generated functions
	    /// </returns>
	    //
        //*********************************************************
        public abstract string GenerateCardSerializationFunctions(string strContainerClassName, RepPP.Card crd, int iIndent);
	}
}
