using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Windows.Forms;
using System.Xml;

namespace RepPP.Toolkit.Window {

    //*********************************************************
    //
	/// <summary>
	/// Add User Interface functionnalities to the listview:
	///     New XP look
	///     Possibility to save and restore its state
	///     Optional support for auto-sorting
	///     Support for single selection extended event
	/// </summary>
	//
    //*********************************************************
	public class ListViewUIExt : ListView, IUIConfiguration {
        /// <summary>
        /// Selected Listview class contains some infos
        /// </summary>
	    protected class SelectionInfo {
            /// <summary>
            /// Current Item
            /// </summary>
	        public ListViewItem         m_itemCurrent;
            /// <summary>
            /// New Item
            /// </summary>
	        public ListViewItem         m_itemNew;
            /// <summary>
            /// Is Selected by user
            /// </summary>
	        public bool                 m_bSelectByUser;
            /// <summary>
            /// Is Reselected old One
            /// </summary>
	        public bool                 m_bReselectOldOne;
            /// <summary>
            /// Could ignore event
            /// </summary>
	        public bool                 m_bIgnoreEvent;
	    };
        /// <summary>
        /// Delegate for the BeforeSelect event
        /// </summary>
        public delegate void                BeforeSelectHandler(object sender, BeforeSelectEventArgs e);
        /// <summary>
        /// Event sent before a line is selected (single-selection)
        /// </summary>
        public event BeforeSelectHandler    BeforeSelect;           // Event sent before an element is selected
        /// <summary>
        /// Delegate for the AfterSelect event
        /// </summary>
        public delegate void                AfterSelectHandler(object sender, AfterSelectEventArgs e);
        /// <summary>
        /// Event sent after a line is selected (single-selection)
        /// </summary>
        public event AfterSelectHandler     AfterSelect;
        /// <summary>
        /// Last selected item
        /// </summary>
        private ListViewItem                m_itemLastSelected;
        /// <summary>
        /// true for auto-sorting by column click
        /// </summary>
        private bool                        m_bAutoSort;
        /// <summary>
        /// true if ascending sort, false if descending
        /// </summary>
        private bool                        m_bAscending = true;
		/// <summary>
		/// Column used for sorting. -1 for none.
		/// </summary>
		private int                         m_iSortColumn   =  -1;
		/// <summary>
		/// Background color
		/// </summary>
		private Color                       m_colorBkgn     = Color.WhiteSmoke;
		/// <summary>
		/// Determines if the selected column is shaded.
		/// </summary>
		private bool                        m_bSortShaded   = true;
		/// <summary>
		/// true to permit key search in the listview, false to forbid it
		/// </summary>
		private bool                        m_bKeySearch = true;
		/// <summary>
		/// Information about the current select operation
		/// </summary>
		protected SelectionInfo             m_selectionInfo = new SelectionInfo();

        //*********************************************************
        //
        /// <summary>
        /// ctor:                       Class constructor
        /// </summary>
        //
        //*********************************************************
        public ListViewUIExt() {
            MultiSelect                         = false;
            m_itemLastSelected                  = null;
            m_bAutoSort                         = false;
	        m_selectionInfo.m_itemCurrent       = null;
	        m_selectionInfo.m_itemNew           = null;
	        m_selectionInfo.m_bSelectByUser     = false;
	        m_selectionInfo.m_bIgnoreEvent      = false;
	        m_selectionInfo.m_bReselectOldOne   = false;
        }

        //*********************************************************
        //
        /// <summary>
        /// SelectedItem:               Get/set the current selected item
        /// </summary>
        //
        //*********************************************************
		[Description("Currently selected item if multi-select if turn off"), Browsable(false)]
        public virtual ListViewItem SelectedItem {
            get {
                ListViewItem    itemRetVal;

                itemRetVal = (SelectedItems.Count == 0) ? null : SelectedItems[0];
                return(itemRetVal);
            }
            set {
                int iSelectedItemsCount;

                iSelectedItemsCount = SelectedItems.Count;
                for (int i = iSelectedItemsCount - 1; i >= 0; i--){  
                    if (SelectedItems[i] != value) {
                        SelectedItems[i].Selected = false;
                    }
                }
                m_itemLastSelected = null;
                if (value != null)
                {
                    value.Selected = true;
                }
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// AutoSort:                   True to autosort by column click
        /// </summary>
        //
        //*********************************************************
		[Description("Defines if the auto-sort is on"), Category("Behavior")]
        public bool AutoSort {
            get {
                return(m_bAutoSort);
            }
            set {
                m_bAutoSort = value;
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// GetColumnOrder:             Gets a column order
        /// </summary>
        /// <param name="iColIndex">    Column index</param>
        /// <returns>
        /// Column order
        /// </returns>
        //
        //*********************************************************
        public int GetColumnOrder(int iColIndex) {
            Win32.LVCOLUMNW     tCol;
            
            tCol        = new Win32.LVCOLUMNW();
            tCol.mask   = Win32.LVCF.ORDER;
            tCol.iOrder = -1;
            Win32.User32.SendMessage(Handle, Win32.LVM.GETCOLUMNW, iColIndex, ref tCol);
            return(tCol.iOrder);
        }

        //*********************************************************
        //
        /// <summary>
        /// SetColumnOrder:             Gets a column order
        /// </summary>
        /// <param name="iColIndex">    Column index</param>
        /// <param name="iOrder">       Order of the column (0..n)</param>
        //
        //*********************************************************
        public void SetColumnOrder(int iColIndex, int iOrder) {
            Win32.LVCOLUMNW     tCol;
            
            tCol        = new Win32.LVCOLUMNW();
            tCol.mask   = Win32.LVCF.ORDER;
            tCol.iOrder = iOrder;
            Win32.User32.SendMessage(Handle, Win32.LVM.SETCOLUMNW, iColIndex, ref tCol);
        }

        //*********************************************************
        //
        /// <summary>
        /// LoadState:                  Implement IUIConfiguration LoadState
        /// </summary>
        /// <param name="helper">       UIConfiguration helper</param>
        /// <param name="node">         Node in which to get the information</param>
        //
        //*********************************************************
        public virtual void LoadState(UIConfigurationHelper helper, XmlNode node) {
            string  strVersion;
            int     iColSize;
            int     iColOrder;
            int     iSortedColumn;
            bool    bAscending;
            bool    bRefresh;
            XmlNode nodeCol;
            
            strVersion      = helper.GetNodeAttribute(node, "VERSION");
            if (strVersion != null && strVersion == "1") {
                iSortedColumn = helper.GetNodeAttribute(node, "SORTEDCOLUMN", -1);
                bAscending    = helper.GetNodeAttribute(node, "ASCENDING", true);
                bRefresh      = false;
                SortColumn(iSortedColumn, bAscending);
                foreach (ColumnHeader col in Columns) {
                    nodeCol    = helper.FindNode(node, "COL" + col.Index.ToString());
                    iColSize   = helper.GetNodeAttribute(nodeCol, "WIDTH", -1);
                    iColOrder  = helper.GetNodeAttribute(nodeCol, "ORDER", -1);
                    if (iColSize != -1) {
                        col.Width = iColSize;
                    }
                    if (iColOrder != -1 && iColOrder != GetColumnOrder(col.Index)) {
                        SetColumnOrder(col.Index, iColOrder);
                        bRefresh = true;
                    }
                }
                if (bRefresh) {
                    Refresh();
                }
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// SaveState:                  Implement IUIConfiguration SaveState
        /// </summary>
        /// <param name="helper">       UIConfiguration helper</param>
        /// <param name="node">         Node in which to save the information</param>
        //
        //*********************************************************
        public virtual void SaveState(UIConfigurationHelper helper, XmlNode node) {
            XmlNode nodeCol;
            int     iColOrder;
            
            node.RemoveAll();
            helper.SetNodeAttribute(node, "VERSION", "1");
            helper.SetNodeAttribute(node, "SORTEDCOLUMN", m_iSortColumn);
            helper.SetNodeAttribute(node, "ASCENDING", m_bAscending);
            foreach (ColumnHeader col in Columns) {
                nodeCol    = helper.FindNode(node, "COL" + col.Index.ToString());
                iColOrder  = GetColumnOrder(col.Index);
                helper.SetNodeAttribute(nodeCol, "WIDTH", col.Width);
                helper.SetNodeAttribute(nodeCol, "ORDER", iColOrder);
            }
        }

        //*********************************************************
        //
		/// <summary>
		/// SortedColumn:       The column on which sorting takes place
		/// </summary>
		//
        //*********************************************************
		[Description("The column on which sorting takes place."), Category("Behavior"), DefaultValue(-1)]
		public int SortedColumn {
			get {
				return(m_iSortColumn);
			}
		}

        //*********************************************************
        //
		/// <summary>
		/// SortColumn:         The column on which sorting takes place
		/// </summary>
		//
        //*********************************************************
		public void SortColumn(int iCol, bool bAscending) {
		    if (iCol != m_iSortColumn || bAscending != m_bAscending) {
		        DrawSortGlyph(m_iSortColumn, SortOrder.None);
		        m_bAscending = bAscending;
		        if (iCol == -1) {
		            ListViewItemSorter = null;
		        } else {
		            m_iSortColumn = iCol;
		            m_bAscending  = bAscending;
		            if (ListViewItemSorter == null) {
		                ListViewItemSorter = new SCListViewItemComparer(this);
		            }
		            Sort();
		            Refresh();
		            Invalidate();
		        }
		    }
		    DrawSortGlyph(m_iSortColumn, m_bAscending ? SortOrder.Ascending : SortOrder.Descending);
		}

        //*********************************************************
        //
		/// <summary>
		/// Ascending:          Sorting order
		/// </summary>
		//
        //*********************************************************
		[Description("Indicates the manner in which items are to be sorted"), Category("Behavior"), DefaultValue(true)]
		public bool Ascending {
			get {
				return(m_bAscending);
			}
		}

        //*********************************************************
        //
		/// <summary>
		/// ShadeColor:         Color of the shaded column in details mode
		/// </summary>
		//
        //*********************************************************
		[Description("Color of the shaded column in details mode"), Category("Appearance")]
		public Color ShadeColor	{
			get	{
				return(m_colorBkgn);
			}
			set	{
				m_colorBkgn = value;
    		    Refresh();
			}
		}

        //*********************************************************
        //
		/// <summary>
		/// Shaded:             If true, the selected column will be shaded when in detail mode.
		/// </summary>
		//
        //*********************************************************
		[Description("If true, the selected column will be shaded when in detail mode."), Category("Appearance"), DefaultValue(true)]
		public bool Shaded {
			get	{
				return(m_bSortShaded);
			}
			set	{
				m_bSortShaded = value;
				Refresh();
			}
		}

        //*********************************************************
        //
		/// <summary>
		/// KeySearch:          Permit keysearch in the listview
		/// </summary>
		//
        //*********************************************************
		[Description("Permit search in the listview using keystroke"), Category("Behavior")]
		public bool KeySearch {
			get	{
				return(m_bKeySearch);
			}
			set	{
				m_bKeySearch = value;
			}
		}

        //*********************************************************
        //
        /// <summary>
        /// OnBeforeSelect:             Called before selecting an item. This event is triggered
        ///                             only on single selection listview
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //*********************************************************
        protected virtual void OnBeforeSelect(BeforeSelectEventArgs e) {
            if (BeforeSelect != null) {
                BeforeSelect(this, e);
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// OnAfterSelect:              Called after selecting an item. This event is triggered
        ///                             only on single selection listview
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //*********************************************************
        protected virtual void OnAfterSelect(AfterSelectEventArgs e) {
            if (AfterSelect != null) {
                AfterSelect(this, e);
            }
        }

        //*********************************************************
        //
        /// <summary>
        /// OnColumnClick:              Called when a column is clicked
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //*********************************************************
        protected override void OnColumnClick(ColumnClickEventArgs e){
            if (m_bAutoSort) {
                if (m_iSortColumn == e.Column) {
                    SortColumn(e.Column, !m_bAscending);
                } else {
                    SortColumn(e.Column, true);
                }
            }
            base.OnColumnClick(e);
        }

        //*********************************************************
        //
        /// <summary>
        /// DrawSortGlyph:              Draw the sort glyph
        /// </summary>
        /// <param name="iColumn">      Column</param>
        /// <param name="eSortOrder">   Sort order</param>
        //
        //*********************************************************
		private void DrawSortGlyph(int iColumn, SortOrder eSortOrder) {
			IntPtr          hHeader;
			Win32.HDITEM    tItem;
		
			hHeader     = Win32.User32.GetDlgItem(Handle, 0);
			tItem       = new Win32.HDITEM();
			tItem.mask  = Win32.HDI.FORMAT;
			Win32.User32.SendMessage(hHeader, Win32.HDM.GETITEMW, iColumn, ref tItem);
			tItem.fmt  &= Win32.HDF.NOSORT;
			if (eSortOrder == SortOrder.Ascending) {
				tItem.fmt |= Win32.HDF.SORTUP;
			} else if (eSortOrder == SortOrder.Descending) {
				tItem.fmt |= Win32.HDF.SORTDOWN;
			}
			Win32.User32.SendMessage(hHeader, Win32.HDM.SETITEMW, iColumn, ref tItem);
		}

        //*********************************************************
        //
        /// <summary>
        /// NotifyCustomDraw:           Custom draw
        /// </summary>
        /// <param name="tCustomDraw">  Custom draw structure</param>
        //
        //*********************************************************
		private Win32.CDRF NotifyCustomDraw(Win32.NMLVCUSTOMDRAW tCustomDraw) {
			Win32.CDRF  eRetVal = Win32.CDRF.DODEFAULT;
			
			switch((Win32.CDDS)tCustomDraw.nmcd.dwDrawStage) {
			case Win32.CDDS.PREPAINT:
				if (m_bSortShaded && View == View.Details) {
					eRetVal = Win32.CDRF.NOTIFYITEMDRAW;
				}
				break;
			case Win32.CDDS.ITEMPREPAINT:
				eRetVal = Win32.CDRF.NOTIFYSUBITEMDRAW;
				break;
			case Win32.CDDS.SUBITEMPREPAINT:
				if (tCustomDraw.nmcd.dwItemSpec < Items.Count) {
					ListViewItem item = Items[tCustomDraw.nmcd.dwItemSpec];
					if (tCustomDraw.iSubItem == 0) {
						item.BackColor = (tCustomDraw.iSubItem == m_iSortColumn) ? m_colorBkgn : BackColor;
					} else if ((tCustomDraw.iSubItem < item.SubItems.Count) && (item.SubItems[tCustomDraw.iSubItem] != null)) {
						item.SubItems[tCustomDraw.iSubItem].BackColor = (tCustomDraw.iSubItem == m_iSortColumn)? m_colorBkgn : BackColor;
					}
				}
				break;
			}
			return(eRetVal);
		}

        //*********************************************************
        //
        /// <summary>
        /// CheckToReselectOldOne:      Reselect the old items if necessary
        /// </summary>
        //
        //*********************************************************
		private void CheckToReselectOldOne() {
		    if (m_selectionInfo.m_bReselectOldOne) {
		        m_selectionInfo.m_bReselectOldOne = false;
		        if (m_selectionInfo.m_itemCurrent != null) {
    		        m_selectionInfo.m_bIgnoreEvent = true;
	    	        try {
		                SelectedItem = m_selectionInfo.m_itemCurrent;
		            } finally {
		                m_selectionInfo.m_bIgnoreEvent = false;
		            }
		        }
		    }
		}

        //*********************************************************
        //
        /// <summary>
        /// WndProc:                    Intercept user selection before it happens
        /// </summary>
        /// <param name="m">            Message</param>
        //
        //*********************************************************
	    protected override void WndProc(ref Message m) {
	        int                         iCount;
            int                         iInternalItemCount = 0;
	        Keys                        key;
	        BeforeSelectEventArgs       eventArgs;
		    Win32.RECT                  rectWin32;
		    Win32.NMHEADER              tNMHeader;
			Win32.NMLVCUSTOMDRAW        tCustomDraw;
			Win32.NMLISTVIEW            tLV;
		    Rectangle                   rect;
	        bool                        bCancel = false;

	        switch((Win32.WM)m.Msg) {
	        case Win32.WM.KEYDOWN:
	        case Win32.WM.LBUTTONDOWN:
	        case Win32.WM.LBUTTONDBLCLK:
	        case Win32.WM.RBUTTONDOWN:
	        case Win32.WM.RBUTTONDBLCLK:
	        case Win32.WM.MOUSEHOVER:
	            m_selectionInfo.m_bSelectByUser = true;
	            try {
                    base.WndProc(ref m);
                } finally {
                    m_selectionInfo.m_bSelectByUser = false;
                }
                CheckToReselectOldOne();
                bCancel = true;
	            break;
	        case Win32.WM.CHAR:
	            key         = (Keys)m.WParam.ToInt32();
	            iCount      = Items.Count;
	            if (iCount != 0 && FocusedItem != null) {
	                switch(key & ~(Keys.Control | Keys.Shift)) {
	                case Keys.Space:
	                    break;
	                default:    // Forbid search in listview
	                    if (!KeySearch) {
	                        bCancel = true;
	                    }
	                    break;
	                }
	            }
	            if (!bCancel) {
	                m_selectionInfo.m_bSelectByUser = true;
	                try {
                        base.WndProc(ref m);
                    } finally {
                        m_selectionInfo.m_bSelectByUser = false;
                    }
                    CheckToReselectOldOne();
                    bCancel = true;
	            }
	            break;
			case Win32.WM.ERASEBKGND:
				// shade background if in details mode
				if (m_bSortShaded && (View == View.Details) && (m_iSortColumn>=0)) {
					rectWin32 = new Win32.RECT();
					Win32.User32.SendMessage(Win32.User32.GetDlgItem(Handle, 0),
					                         Win32.HDM.GETITEMRECT,
					                         m_iSortColumn,
					                         ref rectWin32);
					rect = new Rectangle(rectWin32.left,
					                     Top, 
					                     (rectWin32.right - rectWin32.left),
					                     Height);
                    if (Items.Count > 0) {
                        iInternalItemCount = Win32.User32.SendMessage(Win32.User32.GetDlgItem(Handle, 0), 
                                                 (int)Win32.LVM.GETITEMCOUNT, 
                                                 0, 
                                                 0);
                    }
                    // shift-x with bounds of first item (may be negative if y-scroll)
					rect.X = rect.X + ((Items.Count > 0 && iInternalItemCount > 0) ? GetItemRect(0).X : 0);

					// and fill using the selected background color
					using (Brush brush = new SolidBrush(m_colorBkgn)) {
						Graphics.FromHdc(m.WParam).FillRectangle(brush, rect);
					}
				}
			    break;
			case (Win32.WM)Win32.OCM.NOTIFY:
				tNMHeader = (Win32.NMHEADER)m.GetLParam(typeof(Win32.NMHEADER));
				if (tNMHeader.hdr.hwndFrom == Handle) {
				    switch(tNMHeader.hdr.code) {
				    case (int)Win32.NM.CUSTOMDRAW:
					    tCustomDraw = (Win32.NMLVCUSTOMDRAW)m.GetLParam(typeof(Win32.NMLVCUSTOMDRAW));
					    m.Result    = (IntPtr)NotifyCustomDraw(tCustomDraw);
					    break;
					case (int)Win32.LVN.ITEMCHANGING:
					    if (m_selectionInfo.m_bSelectByUser) {
					        tLV = (Win32.NMLISTVIEW)m.GetLParam(typeof(Win32.NMLISTVIEW));
					        if ((tLV.uNewState & Win32.LVIS.SELECTED) == 0 &&
					            (tLV.uOldState & Win32.LVIS.SELECTED) != 0) {
					            m_selectionInfo.m_itemCurrent = (SelectedItems.Count == 0) ? null : SelectedItems[0];
					        } else if ((tLV.uNewState & Win32.LVIS.SELECTED) != 0 &&
					                   (tLV.uOldState & Win32.LVIS.SELECTED) == 0 &&
					                   SelectedItems.Count == 0) {
					            m_selectionInfo.m_itemNew = Items[tLV.iItem];
    	                        if (!MultiSelect && m_selectionInfo.m_itemNew != m_selectionInfo.m_itemCurrent) {
    	                            eventArgs = new BeforeSelectEventArgs(m_selectionInfo.m_itemCurrent,
    	                                                                  m_selectionInfo.m_itemNew,
                                                                          m_selectionInfo.m_bReselectOldOne);
	                                OnBeforeSelect(eventArgs);
	                                bCancel = eventArgs.Cancel;
        	                        if (bCancel) {
    	        				        m.Result                          = (IntPtr)1;
    	        				        m_selectionInfo.m_bReselectOldOne = true;
	                                }
	                            }
	                        }
					    }
					    break;
					default:
					    break;
					}
				}
			    break;
			}
	        if (!bCancel) {
	            base.WndProc(ref m);
	        }
	    }

        //*********************************************************
        //
        /// <summary>
        /// OnSelectedIndexChanged:     Called when a listview item is selected
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //*********************************************************
	    protected override void OnSelectedIndexChanged(System.EventArgs e) {
	        ListViewItem    item;
	        
	        base.OnSelectedIndexChanged(e);
	        if (!m_selectionInfo.m_bIgnoreEvent) {
                if (!MultiSelect || SelectedItems.Count > 0) {
	                item = SelectedItem;
                    if (item != null && item != m_itemLastSelected && !SelectedItems.Contains(m_itemLastSelected)) {
	                    m_itemLastSelected = item;
    	                OnAfterSelect(new AfterSelectEventArgs(item));
    	            }
                }
	        }
	    }

        //*********************************************************	    
        /// <summary>
        /// Implement a sorter for the listview
        /// </summary>
        //*********************************************************	    
        private class SCListViewItemComparer : IComparer {
            /// <summary>
            /// Father listview
            /// </summary>
            private ListViewUIExt       m_father;

            //*********************************************************	    
            //
            /// <summary>
            /// ctor:                       Class constructor
            /// </summary>
            /// <param name="father">       Father listview</param>
            //
            //*********************************************************	    
            public SCListViewItemComparer(ListViewUIExt father) {
                m_father = father;
            }

            //*********************************************************	    
            //
            /// <summary>
            /// Compare:                    Method comparing values
            /// </summary>
            /// <param name="x">            First value to compare</param>
            /// <param name="y">            Second value to compare</param>
            /// <returns>
            /// -1 for x .lt. y, 0 for x = y, 1 for x .gt. y
            /// </returns>
            //
            //*********************************************************	    
            public int Compare(object x, object y) {
                int             iRetVal;
                ListViewItem    item1;
                ListViewItem    item2;
                bool            bAscending;
                int             iSortColumn;
                
                iSortColumn = m_father.SortedColumn;
                bAscending  = m_father.Ascending;
                item1       = (ListViewItem)x;
                item2       = (ListViewItem)y;
                if (iSortColumn >= item1.SubItems.Count ||
                    iSortColumn >= item2.SubItems.Count) {
                    iRetVal = String.Compare(item1.Text, item2.Text);
                } else {
                    iRetVal = String.Compare(item1.SubItems[iSortColumn].Text, item2.SubItems[iSortColumn].Text);
                }
                if (!bAscending) {
                    iRetVal = -iRetVal;
                }
                return(iRetVal);
            }
        }

        //*********************************************************	    
        /// <summary>
        /// Implement arguments for the SelectingItem event.
        /// </summary>
        //*********************************************************	    
        public class BeforeSelectEventArgs : System.EventArgs {
            private ListViewItem    m_itemOld;
            private ListViewItem    m_itemNew;
            private bool            m_bCancel;
            private bool            m_bReselect;

            //*********************************************************
            //
    	    /// <summary>
    	    /// ctor:                       Class constructor
    	    /// </summary>
    	    /// <param name="itemOld">      Previously selected item</param>
    	    /// <param name="itemNew">      Item to select</param>
    	    /// <param name="bReselect">    Indicates if the item is reselectd</param>
    	    //
            //*********************************************************
            public BeforeSelectEventArgs(ListViewItem itemOld, ListViewItem itemNew, bool bReselect) {
                m_itemOld   = itemOld;
                m_itemNew   = itemNew;
                m_bCancel   = false;
                m_bReselect = bReselect;
            }

            //*********************************************************
            //
    	    /// <summary>
    	    /// OldItem:                Previously selected item
    	    /// </summary>
    	    //
            //*********************************************************
            public ListViewItem OldItem {
                get {
                    return(m_itemOld);
                }
            }

            //*********************************************************
            //
    	    /// <summary>
    	    /// NewItem:                New item to select
    	    /// </summary>
    	    //
            //*********************************************************
            public ListViewItem NewItem {
                get {
                    return(m_itemNew);
                }
            }

            //*********************************************************
            //
    	    /// <summary>
    	    /// Cancel:                 true to cancel the selection of the new item
    	    /// </summary>
    	    //
            //*********************************************************
            public bool Cancel {
                get {
                    return(m_bCancel);
                }
                set {
                    m_bCancel = value;
                }
            }

            //*********************************************************
            //
    	    /// <summary>
    	    /// Reselect:                 Indicates if the item is reselected whene the selection
    	    ///                           is canceled
    	    /// </summary>
    	    //
            //*********************************************************
            public bool Reselect {
                get {
                    return(m_bReselect);
                }
            }
        } // Class BeforeSelectEventArgs

        //*********************************************************
        /// <summary>
        /// Implements the arguments for the AfterSelect event
        /// </summary>
        //*********************************************************
        public class AfterSelectEventArgs : System.EventArgs {
            /// <summary>
            /// Item which has been selected
            /// </summary>
            private ListViewItem    m_item;

            //*********************************************************
            //
    	    /// <summary>
    	    /// ctor:                   Class constructor
    	    /// </summary>
    	    /// <param name="item">     Item selected</param>
    	    //
            //*********************************************************
            public AfterSelectEventArgs(ListViewItem item) {                
                m_item    = item;
            }

            //*********************************************************
            //
    	    /// <summary>
    	    /// Item:                   Item selected
    	    /// </summary>
            //
            //*********************************************************
            public ListViewItem Item {
                get {
                    return(m_item);
                }
            }
        } // Class AfterSelectEventArgs
	} // Class SCListViewUIExt
} // namespace
