using System;
using System.Collections;
using System.Windows.Forms;

namespace RepPP.Toolkit.Window {
    //*********************************************************	    
    //
    /// <summary>
    /// Implement the base class of a tree node attached to a TreeViewCInst 
    /// </summary>
    // 
    //*********************************************************	    
    public class TreeNodeGInst : TreeNodeBase {
        /// <summary>Internal class use to keep information about recursion</summary>
        private class SCRecInfo {
            /// <summary>Recursive group or null if none</summary>
            public RepPP.Group                              m_grpRecursive;
            /// <summary>WHERE clause to make the recursion</summary>
            public string                                   m_strRecWhereClause;
            /// <summary>ORDER BY to order the row of a recursive call</summary>
            public string                                   m_strOrderBy;
            /// <summary>SELECT sql command for 2 levels recursion</summary>
            public string                                   m_strSelectFor2Level;
            /// <summary>Array of child field name if recursion on 2 levels</summary>
            public string[]                                 m_arrChildFldName;
            /// <summary>Cardinstance created if this node is recursive</summary>
            public RepPP.CardInstance                       m_cinstRec;
        };

        private static Hashtable                        m_hashGroupInfo;
        /// <summary>Group instance attached to this node.</summary>
        private RepPP.GroupInstance                     m_ginst;
        /// <summary>Bookmark uniquely identifying the line attached to this node.</summary>
        private int                                     m_iLineBmk;
        /// <summary>Information about the recursion if this node is recursive (null if not)</summary>
        private TreeNodeGInst.SCRecInfo                 m_recInfo;

        //*********************************************************	    
        //
        /// <summary>
        /// ctor:                           Class constructor
        /// </summary>
        /// <param name="treeFather">       Father treeview</param>
        /// <param name="ginst">            Group instance to attach to the node</param>
        /// <param name="iLineIndex">       Line index</param>
        /// <param name="childNodeFactory"> Node factory used to create child. Can be NULL</param>
        //
        //*********************************************************	    
        public TreeNodeGInst(TreeViewCInst                            treeFather,
                             RepPP.GroupInstance                      ginst,
                             int                                      iLineIndex,
                             TreeNodeGInst.IGInstChildNodeFactory     childNodeFactory) : base(childNodeFactory)
        {
            m_ginst             = ginst;
            m_iLineBmk          = ginst.LineBookmark(iLineIndex);
            m_recInfo           = null;
            Init(treeFather, ginst.ChildGroupInstances.Count != 0); 
        }

        //*********************************************************	    
        //
        /// <summary>
        /// ctor:                           Class constructor
        /// </summary>
        //
        //*********************************************************	    
        public TreeNodeGInst() 
        {
            m_ginst             = null;
            m_iLineBmk          = 0;
            m_recInfo           = null;
        }

        //*********************************************************	    
        //
        /// <summary>
        /// Dispose:                Free the node
        /// </summary>
        //
        //*********************************************************	    
        public override void Dispose()
        {
            if (m_recInfo != null) {
                if (m_recInfo.m_cinstRec != null) {
                    m_recInfo.m_cinstRec.Delete();
                    m_recInfo.m_cinstRec = null;
                }
                m_recInfo = null;
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// Clone:                      Clone the node and all its descendant
        /// </summary>
        /// <returns>
        /// Node key
        ///</returns>
        //
        //*********************************************************	    
        public override object Clone() {
            TreeNodeGInst   nodeRetVal;
            SCRecInfo       recInfo;
            CardInstance    ciNew;

            nodeRetVal                          = base.Clone() as TreeNodeGInst;
            nodeRetVal.m_ginst                  = m_ginst;
            nodeRetVal.m_iLineBmk               = m_iLineBmk;
            if (m_recInfo == null) {
                nodeRetVal.m_recInfo            = null;
            } else {
                recInfo                         = new SCRecInfo();
                recInfo.m_grpRecursive          = m_recInfo.m_grpRecursive;
                recInfo.m_strRecWhereClause     = m_recInfo.m_strRecWhereClause;
                recInfo.m_strOrderBy            = m_recInfo.m_strOrderBy;
                recInfo.m_strSelectFor2Level    = m_recInfo.m_strSelectFor2Level;
                recInfo.m_arrChildFldName       = m_recInfo.m_arrChildFldName;
                if (m_recInfo.m_cinstRec == null) {
                    recInfo.m_cinstRec  = null;
                } else {
                    ciNew               = m_recInfo.m_cinstRec.Card.CardInstances.Add();
                    ciNew.RootGroupInstance.CopyFrom(m_recInfo.m_cinstRec.RootGroupInstance, true);
                    recInfo.m_cinstRec  = ciNew;
                }
            }
            return(nodeRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SpiceSelectCmd: Spice the SELECT cmd to execute a join between the child and father table. This
        ///                 spicing of the SELECT cmd need to be done for recursion using 2 groups
        /// </summary>
        //
        // Transform SELECT P1,P2,S1,S2 FROM T1 into
        //           SELECT A.P1,A.P2,S1,S2 FROM T1 A,T2 B WHERE A.P1=B.X1 AND A.P2=B.X2 AND B.P1=:T1.P1 AND B.P2=:T2.P2
        //           Where  Pn is primary key of the root card group
        //                  Xn is the child foreign key
        //                  T2 is the table name of the specified card group...
        //
        //*********************************************************	    
        private void SpiceSelectCmd() {
            RepPP.Group     grp;
            RepPP.Fields    fldsT1;
            RepPP.Fields    fldsT2;
            RepPP.Field     fldT2;
            RepPP.Tool      tool;
            string          strWhere;
            string          strWhere2;
            string          strSqlCmd;
            string          strOriOrderBy;
            string          strTableName = "";
            int             iPKCount;
            int             iPos;

            tool            = m_ginst.Application.Tool;            
            grp             = m_ginst.Group;
            fldsT1          = grp.Fields;
            fldsT2          = m_recInfo.m_grpRecursive.Fields;
            iPKCount        = 0;
            strSqlCmd       = m_ginst.CardGroup.SelectSqlCommand;
            strWhere2       = "";
            strOriOrderBy   = tool.GetSqlOrderBy(strSqlCmd);
            foreach (RepPP.Field fld in fldsT1) {
                if (fld.IsPrimaryKey) {
                    iPKCount++;
                    strWhere2 += " AND B." + fld.ColName + "=:" + grp.Name + "." + fld.Name;
                    if (StringUtility.StrFindKeyword(strSqlCmd,
                                                     0,
                                                     out iPos,
                                                     fld.ColName,
                                                     StringUtility.SearchTypeE.SK_NON_ALNUM_AFTER  |
                                                     StringUtility.SearchTypeE.SK_NON_ALNUM_BEFORE |
                                                     StringUtility.SearchTypeE.SK_IGNORE_CASE      |
                                                     StringUtility.SearchTypeE.SK_NOT_IN_SIMPLE_QUOTE) != -1) {
                                                     strSqlCmd = strSqlCmd.Substring(0, iPos) + "A." + strSqlCmd.Substring(iPos);
                    }
                    if (StringUtility.StrFindKeyword(strOriOrderBy,
                                                     0,
                                                     out iPos,
                                                     fld.ColName,
                                                     StringUtility.SearchTypeE.SK_NON_ALNUM_AFTER  |
                                                     StringUtility.SearchTypeE.SK_NON_ALNUM_BEFORE |
                                                     StringUtility.SearchTypeE.SK_IGNORE_CASE      |
                                                     StringUtility.SearchTypeE.SK_NOT_IN_SIMPLE_QUOTE) != -1) {
                                                     strOriOrderBy = strOriOrderBy.Substring(0, iPos) + "A." + strOriOrderBy.Substring(iPos);
                    }
                }
            }
            if (iPKCount != m_recInfo.m_arrChildFldName.Length) {
                throw new System.ApplicationException("Bad list of child fields. Array must match the primary key");
            }
            strWhere = " WHERE ";
            iPKCount = 0;
            foreach (RepPP.Field fldT1 in fldsT1) {
                if (fldT1.IsPrimaryKey) {
                    fldT2 = fldsT2[m_recInfo.m_arrChildFldName[iPKCount]];
                    if (fldT2 == null) {
                        throw new System.ApplicationException("Field '" + m_recInfo.m_arrChildFldName[iPKCount] +
                                                              "' not found!");
                    }
                    if (iPKCount != 0) {
                        strWhere += " AND ";
                    }
                    strWhere += "A." + fldT1.ColName + "=B." + fldT2.ColName;
                    strTableName = fldT2.TableName;
                    iPKCount++;
                }
            }
            m_recInfo.m_strSelectFor2Level = tool.SetSqlWhere(strSqlCmd, "A, " + strTableName + " B " + strWhere + strWhere2);
            m_recInfo.m_strSelectFor2Level = tool.SetSqlOrderBy(m_recInfo.m_strSelectFor2Level, (m_recInfo.m_strOrderBy != null) ? m_recInfo.m_strOrderBy : strOriOrderBy);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SetRecursive:                   Transform this node into a recursive one
        /// </summary>
        /// <param name="grpRecursive">     Group which make the recursion. NULL for the root.</param>
        /// <param name="strRecWhereClause">WHERE clause use to make the recursion</param>
        /// <param name="strOrderBy">       ORDER BY clause or null if none</param>
        /// <param name="arrChildFldName">      List of child field which make the link</param>
        //
        //*********************************************************	    
        private void SetRecursive(RepPP.Group    grpRecursive,
                                  string         strRecWhereClause,
                                  string         strOrderBy,
                                  string[]       arrChildFldName) {
            if (m_recInfo != null) {
                throw new System.ApplicationException("The recursion is already set for this node!");
            }
            try {
                if (m_ginst.CardInstance == null) {
                    throw new System.ApplicationException("The attached group instance of a node must be part of a card instance to be able to be recursive");
                } else if (m_ginst.CardInstance.RootGroupInstance != m_ginst) {
                    throw new System.ApplicationException("Only a node attached to the root group instance of a card instance can be set as recursive");
                }
                m_recInfo                      = new SCRecInfo();
                m_recInfo.m_strSelectFor2Level = "";
                m_recInfo.m_strRecWhereClause  = strRecWhereClause;
                m_recInfo.m_strOrderBy         = strOrderBy;
                m_recInfo.m_arrChildFldName    = arrChildFldName;
                m_recInfo.m_cinstRec           = null;
                if (grpRecursive == null || grpRecursive == m_ginst.Group) {
                    m_recInfo.m_grpRecursive = m_ginst.Group;
                } else {
                    foreach (RepPP.CardGroup cgrp in m_ginst.CardInstance.Card.ChildCardGroups[0].ChildCardGroups) {
                        if (cgrp.Group == grpRecursive) {
                            m_recInfo.m_grpRecursive = grpRecursive;
                            break;
                        }
                    }
                    if (m_recInfo.m_grpRecursive == null) {
                        throw new System.ApplicationException("The group making the recursion must be the root group or a direct descendant");
                    }
                    if (arrChildFldName == null) {
                        throw new System.ApplicationException("You must specify a list of fields which contains the link in the child group to the father primary key");
                    }
                    SpiceSelectCmd();
                }
            } catch {
                m_recInfo = null;
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SetRecursive1Level:             Transform this node into a recursive one. The recursion must
        ///                                 occurs within the root group.
        /// </summary>
        /// <param name="strRecWhereClause">WHERE clause use to make the recursion</param>
        /// <param name="strOrderBy">       ORDER BY clause or null if none</param>
        //
        //*********************************************************	    
        public void SetRecursive1Level(string   strRecWhereClause,
                                       string   strOrderBy) {

            SetRecursive(null,
                         strRecWhereClause,
                         strOrderBy,
                         null);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SetRecursive2Level:             Transform this node into a recursive one. The recursion is done
        ///                                 between the root group and the specified one (which must be a direct
        ///                                 descendant of the root one).
        /// </summary>
        /// <param name="grpRecursive">     Group which make the recursion.</param>
        /// <param name="strOrderBy">       ORDER BY clause or null if none</param>
        /// <param name="arrChildFldName">  List of child field which make the link</param>
        //
        //*********************************************************	    
        public void SetRecursive2Level(RepPP.Group    grpRecursive,
                                       string         strOrderBy,
                                       string[]       arrChildFldName) {
            SetRecursive(grpRecursive,
                         "",
                         strOrderBy,
                         arrChildFldName);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GroupInstance:      Group instance attached to this node
        /// </summary>
        //
        //*********************************************************	    
        public override RepPP.GroupInstance GroupInstance {
            get {
                return(m_ginst);
            }
        }

        //******************************************************************
        //
        /// <summary>
        /// ReassociateGInst        Reassociate the node to a new group instance.
        /// </summary>
        /// <param name="giNew">    New group instance to associate</param>
        /// <param name="iNewLine"> New line of the item (-1 if didn't change)</param>
        //
        //******************************************************************
        public virtual void ReassociateGInst(GroupInstance giNew, int iNewLine) {
            GroupInstance       giOld;
            GroupInstance       giChildOld;
            GroupInstance       giChildNew;
            TreeNodeGInst       nodeChild;
            int                 iLine;

            giOld   = GroupInstance;
            iLine   = (iNewLine != -1) ? iNewLine : giOld.LineFromBookmark(m_iLineBmk);
            if (giOld.ChildGroupInstances.Count != 0) {
                foreach (TreeNode nodeTmp in Nodes) {
                    nodeChild = nodeTmp as TreeNodeGInst;
                    if (nodeChild != null) {
                        giChildOld = nodeChild.GroupInstance;
                        if (giChildOld != null && giChildOld.FatherGroupInstance == giOld) {
                            giChildNew = giNew.GetChild(giChildOld.CardGroup.ChildId, iLine, false, false);
                            nodeChild.ReassociateGInst(giChildNew, -1);
                        }
                    }
                }
            }
            m_iLineBmk = giNew.LineBookmark(iLine);
            m_ginst    = giNew;
        }

        //*********************************************************	    
        //
        /// <summary>
        /// FindChildGroupInstance:     Find a child group instance if any
        /// </summary>
        /// <param name="grp">          Group</param>
        /// <returns>
        /// null or the specified group instance
        /// </returns>
        //
        //*********************************************************	    
        public override RepPP.GroupInstance FindChildGroupInstance(RepPP.Group grp)
        {
            RepPP.GroupInstance ginstRetVal = null;
            
            if (m_ginst != null) {
                ginstRetVal = m_ginst.ChildGroupInstances.Find(grp.Name);
            }
            return(ginstRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// LineBookmark:       Bookmark identifying the group instance line attached to this node.
        /// </summary>
        //
        //*********************************************************	    
        public int LineBookmark {
            get {
                return(m_iLineBmk);
            }
            set {
                m_iLineBmk = value;
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// IsRecursive:            true if the node is recursive, no if not.
        /// </summary>
        //
        //*********************************************************	    
        public override bool IsRecursive
        {
            get {
                return(m_recInfo != null);
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        ///  SpiceSelectCommand     Returned the spiced SQL command if any
        /// </summary>
        //
        //*********************************************************	    
        public string SpiceSelectCommand
        {
            get {
                return((m_recInfo != null) ? m_recInfo.m_strSelectFor2Level : null);
            }
            set {
                if (m_recInfo == null) {
                    throw new System.ApplicationException("You must first set the node recursive!");
                }
                m_recInfo.m_strSelectFor2Level = value;
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// RecCardInstance:        Returned the card instance created by this node to hold
        ///                         the recursive children group instance.
        /// </summary>
        //
        //*********************************************************	    
        public RepPP.CardInstance RecCardInstance
        {
            get {
                CardInstance    ciRetVal;

                if (m_recInfo == null) {
                    ciRetVal = null;
                } else {
                    if (m_recInfo.m_cinstRec == null && m_ginst != null) {
                        m_recInfo.m_cinstRec = m_ginst.CardInstance.Card.CardInstances.Add();
                    }
                    ciRetVal = m_recInfo.m_cinstRec;
                }
                return(ciRetVal);
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// CreateNode:                     Creates a node
        /// </summary>
        /// <param name="treeFather">       Father treeview</param>
        /// <param name="ginst">            Group instance</param>
        /// <param name="iLineIndex">       Line index</param>
        /// <param name="childNodeFactory"> Node factory used to create child. Can be NULL</param>
        //
        //*********************************************************	    
        public static TreeNodeGInst CreateNode(TreeViewCInst                           treeFather,
                                               RepPP.GroupInstance                     ginst,
                                               int                                     iLineIndex,
                                               TreeNodeGInst.IGInstChildNodeFactory    childNodeFactory)
        {
            return((childNodeFactory == null) ? new TreeNodeGInst(treeFather, ginst, iLineIndex, null) :
                                                    childNodeFactory.CreateGInstNode(treeFather,
                                                                                     ginst,
                                                                                     iLineIndex,
                                                                                     childNodeFactory));
        }                                                    

        //*********************************************************	    
        //
        /// <summary>
        /// FillNodes:                      Fill the nodes with the content of the group instance
        /// </summary>
        /// <param name="treeFather">       Father treeview</param>
        /// <param name="nodes">            Nodes collections</param>
        /// <param name="ginst">            Group instance</param>
        /// <param name="childNodeFactory"> Node factory used to create child. Can be NULL</param>
        //
        //*********************************************************	    
        public static void FillNodes(TreeViewCInst                           treeFather,
                                     TreeNodeCollection                      nodes,
                                     RepPP.GroupInstance                     ginst,
                                     TreeNodeGInst.IGInstChildNodeFactory    childNodeFactory)
        {
            int             iLineCount;
            int             iLineIndex;
            int             iIndex;
            TreeNodeBase[]  nodesToAdd;
            
            treeFather.BeginUpdate();
            try {
                iLineCount = ginst.LineCount;
                nodesToAdd = new TreeNodeBase[ginst.UndelLineCount];
                iIndex     = 0;
                for (iLineIndex = 0; iLineIndex < iLineCount; iLineIndex++) {
                    if (ginst.ThisLineState(iLineIndex) != LineState.sdLineDeleted) {
                        nodesToAdd[iIndex++] = CreateNode(treeFather, ginst, iLineIndex, childNodeFactory);
                    }
                }
                nodes.AddRange(nodesToAdd);
            } finally {
                treeFather.EndUpdate();
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// FillOneChildNodes:              Fill the nodes with the content of the group instance
        /// </summary>
        /// <param name="treeFather">       Father treeview</param>
        /// <param name="ginst">            Group instance</param>
        /// <param name="childNodeFactory"> Node factory used to create child. Can be NULL</param>
        //
        //*********************************************************	    
        public virtual void FillOneChildNodes(TreeViewCInst                           treeFather,
                                              RepPP.GroupInstance                     ginst,
                                              TreeNodeGInst.IGInstChildNodeFactory    childNodeFactory)
        {
            FillNodes(treeFather, Nodes, ginst, childNodeFactory);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// ReadChildren:           Read node's children
        /// </summary>
        //
        //*********************************************************	    
        public override void ReadChildren()
        {
            RepPP.CardInstance                      ciRec;
            RepPP.GroupInstance                     ginstChild;
            int                                     iAttachedLine;
            TreeViewCInst.BeforeChildReadEventArgs  beforeArg;
            TreeViewCInst.AfterChildReadEventArgs   afterArg;
            string                                  strOldSelectCmd;
            
            if (!IsChildrenRead) {
                IsChildrenRead  = true;
                // Remove the 'Dummy' node created by the base class so the '+' appear
                Nodes.Clear();
                beforeArg       = new TreeViewCInst.BeforeChildReadEventArgs(this);
                OnBeforeChildRead(beforeArg);
                if (!beforeArg.Cancel) {
                    iAttachedLine   = m_ginst.LineFromBookmark(m_iLineBmk);
                    if (iAttachedLine != -1) {
                        m_ginst.Group.SaveActiveGroupInstance();
                        m_ginst.SaveLinePos();
                        try {
                            m_ginst.Select(false, false);
                            m_ginst.SelectLine(iAttachedLine, false, false);
                            if (m_recInfo != null) {
                                // Recursive 1 level: the group point to itself
                                strOldSelectCmd = m_ginst.CardGroup.SelectSqlCommand;
                                try {
                                    if (m_recInfo.m_grpRecursive == m_ginst.Group) {
                                        m_ginst.CardGroup.SetSelectSqlWhere(m_recInfo.m_strRecWhereClause);
                                    } else {
                                        m_ginst.CardGroup.SelectSqlCommand = m_recInfo.m_strSelectFor2Level;
                                    }
                                    if (m_recInfo.m_strOrderBy != null) {
                                        m_ginst.CardGroup.SetSelectSqlOrderBy(m_recInfo.m_strOrderBy);
                                    }
                                    ciRec = RecCardInstance;
                                    m_ginst.Select(false, false);
                                    m_recInfo.m_cinstRec.ReadFromDb();
                                } finally {
                                    m_ginst.CardGroup.SelectSqlCommand = strOldSelectCmd;
                                }
                                FillOneChildNodes(TreeView as TreeViewCInst, m_recInfo.m_cinstRec.RootGroupInstance, ChildNodeFactory);
                            }
                            foreach (RepPP.CardGroup cgrp in m_ginst.CardGroup.ChildCardGroups) {
                                if (m_recInfo == null || cgrp.Group != m_recInfo.m_grpRecursive) {
                                    ginstChild = m_ginst.ChildGroupInstances[cgrp.ChildId];
                                    FillOneChildNodes(TreeView as TreeViewCInst, ginstChild, ChildNodeFactory);
                                }
                            }
                        } finally {
                            m_ginst.RestoreLinePos(false);
                            m_ginst.Group.RestoreActiveGroupInstance(false);
                        }
                    }
                    afterArg = new TreeViewCInst.AfterChildReadEventArgs(this);
                    OnAfterChildRead(afterArg);
                }
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SetNodeInfo:                Set the node information
        /// </summary>
        /// <param name="treeFather">   Father treeview</param>
        //
        //*********************************************************	    
        protected override void SetNodeInfo(TreeViewCInst treeFather) {
            RepPP.Field[]   arrFlds;
            ArrayList       arrList;
            RepPP.Group     grp;
            RepPP.Fields    flds;
            int             iLineIndex;
            bool            bPKInList;
            string          strUserText = null;
            
            if (m_hashGroupInfo == null) {
                m_hashGroupInfo = new Hashtable(10);
            }
            if (m_ginst != null) {
                grp = m_ginst.Group;
                if (m_hashGroupInfo.ContainsKey(grp)) {
                    arrFlds = m_hashGroupInfo[grp] as RepPP.Field[];
                } else {
                    arrList     = new ArrayList();
                    flds        = m_ginst.Group.Fields;
                    bPKInList   = false;
                    foreach(RepPP.Field fld in m_ginst.Group.Fields) {
                        if (fld.IsPrimaryKey && arrList.Count == 0) {
                            arrList.Add(fld);
                            bPKInList = true;
                        }
                        if (fld.IsUserFlagOn) {
                            if (bPKInList) {
                                arrList.Clear();
                                bPKInList = false;
                            }
                            arrList.Add(fld);
                        }
                    }
                    if (arrList.Count == 0) {
                        arrList.Add(flds[0]);
                    }
                    arrFlds = new RepPP.Field[arrList.Count];
                    arrList.CopyTo(arrFlds);
                    m_hashGroupInfo.Add(grp, arrFlds);
                }
                iLineIndex  = m_ginst.LineFromBookmark(m_iLineBmk);
                strUserText = null;
                foreach (RepPP.Field fld in arrFlds) {
                    if (strUserText == null) {
                        strUserText = m_ginst.GetFieldValue(fld, iLineIndex, true);
                    } else {
                        strUserText += " " + m_ginst.GetFieldValue(fld, iLineIndex, true);
                    }
                }
                Text    = strUserText;
                Tooltip = strUserText;
                base.SetNodeInfo(treeFather);
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GetNodeKey:                 Gets the node unique identifier.
        /// </summary>
        /// <returns>
        /// Key uniquely identifying the node
        ///</returns>
        //
        //*********************************************************	    
        public override string GetNodeKey()
        {
            string      strRetVal;
            int         iAttachedLine;
            
            strRetVal     = "GINST:";
            iAttachedLine = m_ginst.LineFromBookmark(m_iLineBmk);
            if (iAttachedLine != -1) {
                m_ginst.Group.SaveActiveGroupInstance();
                m_ginst.SaveLinePos();
                try {
                    m_ginst.Select(false, false);
                    if (iAttachedLine != m_ginst.CurrentLine) {
                        m_ginst.SelectLine(iAttachedLine, false, false);
                    }
                    foreach (RepPP.Field fld in m_ginst.Group.Fields) {
                        if (fld.IsPrimaryKey) {
                            if (strRetVal.Length != 0) {
                                strRetVal += ",";
                            }
                            strRetVal += fld.Group.Name + "." + fld.Name + "," + fld.Value;
                        }
                    }
                } finally {
                    m_ginst.RestoreLinePos(false);
                    m_ginst.Group.RestoreActiveGroupInstance(false);
                }
            }
            return(strRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// SelectAttachedObject:       Select the object attached to the node
        /// </summary>
        //
        //*********************************************************	    
        public override void SelectAttachedObject()
        {
            int             iLineIndex;
            
            base.SelectAttachedObject();
            if (m_ginst != m_ginst.ActiveGroupInstance) {
                m_ginst.Select(false, true);
            }
            if (m_iLineBmk != -1) {
                iLineIndex = m_ginst.LineFromBookmark(m_iLineBmk);
                if (iLineIndex != -1) {
                    m_ginst.SelectLine(iLineIndex, false, true);
                }
            }
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GetSiblingCount:        Gets the number of sibling
        /// </summary>
        /// <returns>
        /// Number of nodes of the same type.
        /// </returns>
        //
        //*********************************************************	    
        public override int GetSiblingCount()
        {
            int                 iRetVal = 0;
            RepPP.GroupInstance ginst;
            TreeNodeGInst       nodeSibling;
            TreeNodeCollection  nodes;
            
            ginst       = GroupInstance;
            nodes       = (Parent == null) ? TreeView.Nodes : Parent.Nodes;
            iRetVal     = 0;
            foreach (TreeNode nodeTmp in nodes) {
                nodeSibling = nodeTmp as TreeNodeGInst;
                if (nodeSibling != null && nodeSibling.GroupInstance == ginst) {
                    iRetVal++;
                }
            }
            return(iRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// GetSiblingPos:          Gets the position in the sibling
        /// </summary>
        /// <returns>
        /// Position
        /// </returns>
        //
        //*********************************************************	    
        public override int GetSiblingPos()
        {
            int                 iRetVal;
            TreeNodeCollection  nodes;
            TreeNodeGInst       nodeSibling;
            RepPP.GroupInstance ginst;
            
            ginst       = GroupInstance;
            nodes       = (Parent == null) ? TreeView.Nodes : Parent.Nodes;
            iRetVal     = 0;
            foreach (TreeNode nodeTmp in nodes) {
                if (nodeTmp == this) {
                    break;
                }
                nodeSibling = nodeTmp as TreeNodeGInst;
                if (nodeSibling != null && nodeSibling.GroupInstance == ginst) {
                    iRetVal++;
                }
            }
            return(iRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// FindSiblingNode:        Find a sibling node using a position
        /// </summary>
        /// <param name="iPos">     Position (0..n)</param>
        /// <returns>
        /// null or the specified Node
        /// </returns>
        //
        //*********************************************************	    
        public override TreeNodeBase FindSiblingNode(int iPos)
        {
            TreeNodeBase        nodeRetVal = null;
            TreeNodeCollection  nodes;
            TreeNodeGInst       nodeSibling;
            RepPP.GroupInstance ginst;
            
            ginst       = GroupInstance;
            nodes       = (Parent == null) ? TreeView.Nodes : Parent.Nodes;
            foreach (TreeNode nodeTmp in nodes) {
                nodeSibling = nodeTmp as TreeNodeGInst;
                if (nodeSibling != null && nodeSibling.GroupInstance == ginst) {
                    if (iPos == 0) {
                        nodeRetVal = nodeSibling;
                        break;
                    } else {
                        iPos--;
                    }
                }
            }
            return(nodeRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// Group instance node factory interface
        /// </summary>
        //
        //*********************************************************	    
        public interface IGInstChildNodeFactory {
            //*********************************************************	    
            //
            /// <summary>
            /// CreateGInstNode:                Creates a group instance node
            /// </summary>
            /// <param name="treeFather">       Father treeview</param>
            /// <param name="ginst">            Group instance attached to the node</param>
            /// <param name="iLineIndex">       Line index</param>
            /// <param name="childNodeFactory"> Factory to use for children</param>
            /// <returns>
            /// The new node
            /// </returns>
            //
            //*********************************************************	    
            TreeNodeGInst CreateGInstNode(TreeViewCInst treeFather, RepPP.GroupInstance ginst, int iLineIndex, IGInstChildNodeFactory childNodeFactory);
        }
    }
}
